/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>

#include "mpm_transport_sharedmem.h"

#define MAX_DEVICE_NAME_LEN 32

static int mpm_transport_get_global_addr (mpm_transport_cfg_t *sp, uint32_t laddr, uint32_t size, uint32_t *gaddr, int *index)
{
	int i;

	for (i = 0; i < sp->num_mmap; i++) {

		if (sp->mmap[i].local_addr <= laddr &&
		    sp->mmap[i].local_addr + sp->mmap[i].length >= laddr + size) {
			*gaddr = sp->mmap[i].global_addr + laddr - sp->mmap[i].local_addr;
			*index = i;
			return 0;
		}
	}

	for (i = 0; i < sp->num_mmap; i++) {	
		if (sp->mmap[i].global_addr <= laddr &&
		    sp->mmap[i].global_addr + sp->mmap[i].length >= laddr + size) {
			*gaddr = laddr;
			*index = i;
			return 0;
		}
	}

	mpm_printf(1, "can't find global address from local address 0x%x size %d in cfg for %s\n",
	       laddr, size, sp->name);
	return -1;
}

int mpm_transport_sharedmem_open(mpm_transport_cfg_t *sp, mpm_transport_open_t *ocfg)
{
	int i, j;
	mpm_transport_sharedmem_t *td;
	char dev_name[MAX_DEVICE_NAME_LEN];

	td = calloc(1, sizeof(mpm_transport_sharedmem_t));
	if (!td) {
		mpm_printf(1, "can't allocate memory for transport data \n");
		return -1;
	}

	for (i = 0; i < sp->num_mmap; i++) {
		/* check if fd for same device name is already opened */
		for (j = 0; j < i; j++) {
			if (strncmp(sp->mmap[i].name, sp->mmap[j].name, MAX_DEVICE_NAME_LEN)) {
				continue;
			}
			td->fd[i] = td->fd[j];
			break;
		}
		if (j < i) {
			break;
		}

		td->fd[i] = open(sp->mmap[i].devicename, (O_RDWR | O_SYNC));
		if (td->fd[i] < 0) {
			mpm_printf(1, "Failed to open %s :Error (%s)",
			       sp->mmap[i].devicename, strerror(errno));
			free(td);
			return -1;
		}
	}

	sp->td = (void *)td;
	return 0;

}

void *mpm_transport_sharedmem_mmap(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, mpm_transport_mmap_t *mcfg)
{
	int index;
	int user_index;
	uint32_t gaddr;
	int offset, pg_offset, mmap_length, min_len;
	int page_size = getpagesize();
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;

	/* Find a free mmap index to use */
	for (user_index = 0; user_index < MPM_MAX_USER_MMAPS; user_index++) {
		if (!td->mmap_user[user_index].addr) {
			break;
		}
	}

	if (user_index >= MPM_MAX_USER_MMAPS) {
		mpm_printf(1, "exceeded maximum user mmap limits (%d)\n",
 		       MPM_MAX_USER_MMAPS);
		return 0;
	}

	if(mpm_transport_get_global_addr(sp, addr, length, &gaddr, &index)) {
		mpm_printf(1, "can't find global address for local address 0x%x\n", addr);
		return 0;
	}

	offset = gaddr;
	pg_offset = offset & (~(page_size - 1));

	min_len = length + offset - pg_offset;
	mmap_length = (min_len & ~(page_size - 1));
	mmap_length = (mmap_length != min_len) ? mmap_length + page_size : mmap_length;

	td->mmap_user[user_index].addr = mmap(NULL, mmap_length, mcfg->mmap_prot, mcfg->mmap_flags,
						td->fd[index], pg_offset);
	if (td->mmap_user[user_index].addr == MAP_FAILED) {
		mpm_printf(1, "can't mmap for the address 0x%x with length 0x%x (err: %s)\n",
 		       addr, length, strerror(errno));
		return 0;
	}
	td->mmap_user[user_index].size = mmap_length;
	td->mmap_user[user_index].addr_usr = td->mmap_user[user_index].addr + (offset - pg_offset);

	return (td->mmap_user[user_index].addr_usr);
}

int mpm_transport_sharedmem_munmap(mpm_transport_cfg_t *sp, void *va, uint32_t length)
{
	int user_index;
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;
	
	for (user_index = 0; user_index < MPM_MAX_USER_MMAPS; user_index++) {
		if (td->mmap_user[user_index].addr_usr == va) {
			break;
		}
	}

	if (user_index >= MPM_MAX_USER_MMAPS) {
		mpm_printf(1, "can't find mapping to unmap\n",
 		       MPM_MAX_USER_MMAPS);
		return -1;
	}

	if (length > td->mmap_user[user_index].size) {
		mpm_printf(1, "invalid length %d, it must be <= %d\n",
 		       length, td->mmap_user[user_index].size);
	}

	munmap(td->mmap_user[user_index].addr, td->mmap_user[user_index].size);

	td->mmap_user[user_index].addr = 0;
	td->mmap_user[user_index].size = 0;
	return 0;
}

static void *mpm_transport_sharedmem_rw_map (mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, int *index)
{
	uint32_t gaddr;
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;

	if(mpm_transport_get_global_addr(sp, addr, length, &gaddr, index)) {
		mpm_printf(1, "can't find global address for local address 0x%x\n", addr);
		return 0;
	}

	if (!td->mmap_rw[*index].addr) {
		td->mmap_rw[*index].addr = mmap(NULL, sp->mmap[*index].length, (PROT_READ|PROT_WRITE),
			    MAP_SHARED, td->fd[*index], sp->mmap[*index].global_addr);
		if (td->mmap_rw[*index].addr == MAP_FAILED) {
			mpm_printf(1, "can't mmap for the address 0x%x with length 0x%x (err: %s)\n",
	 		       addr, length, strerror(errno));
			return 0;
		}

		td->mmap_rw[*index].size = sp->mmap[*index].length;
		pthread_mutex_init(&td->mutex_rw[*index], NULL);
	}

	return (td->mmap_rw[*index].addr + (gaddr - sp->mmap[*index].global_addr));
}

int mpm_transport_sharedmem_read(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, char *buf, mpm_transport_read_t *rcfg)
{
	int index;
	void *mmap_addr;
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;

	mmap_addr = mpm_transport_sharedmem_rw_map (sp, addr, length, &index);

	if (!mmap_addr) {
		mpm_printf(1, "can't map for address 0x%x\n", addr);
		return -1;
	}
	pthread_mutex_lock(&td->mutex_rw[index]);
	memcpy(buf, mmap_addr, length);
	pthread_mutex_unlock(&td->mutex_rw[index]);
	return 0;
}

int mpm_transport_sharedmem_write(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, char *buf, mpm_transport_write_t *cfg)
{
	int index;
	void *mmap_addr;
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;

	mmap_addr = mpm_transport_sharedmem_rw_map (sp, addr, length, &index);
	if (!mmap_addr) {
		mpm_printf(1, "can't map for address 0x%x\n", addr);
		return -1;
	}
	pthread_mutex_lock(&td->mutex_rw[index]);
	memcpy(mmap_addr, buf, length);
	pthread_mutex_unlock(&td->mutex_rw[index]);
	return 0;
}

void mpm_transport_sharedmem_close(mpm_transport_cfg_t *sp)
{
	int i, j;
	mpm_transport_sharedmem_t *td = (mpm_transport_sharedmem_t *) sp->td;

	if (!td) {
		return;
	}

	for (i = 0; i < MPM_MAX_USER_MMAPS; i++) {
		if(td->mmap_user[i].addr) {
			munmap(td->mmap_user[i].addr, td->mmap_user[i].size);
		}
	}
	for (i = 0; i < MPM_MAX_MMAPS; i++) {
		if(td->mmap_rw[i].addr) {
			munmap(td->mmap_rw[i].addr, td->mmap_rw[i].size);
		}
		if (td->fd[i]) {
			for (j = i + 1; j < MPM_MAX_MMAPS; j++) {
				if (td->fd[i] == td->fd[j]) {
					td->fd[j] = 0;
				}
			}
			fsync(td->fd[i]);
			close(td->fd[i]);
		}
		pthread_mutex_destroy(&td->mutex_rw[i]);
	}

	if (td) free(td);
}
